@@ -51,21 +51,22 @@ __attribute__((constructor)) void init(void) {
 }
 */
 import "C"
+
 import (
 	"bufio"
 	"bytes"
 	"context"
+	"embed"
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"os"
 	"os/signal"
 	"path/filepath"
 	"strings"
 	"time"
 
-	lxd "github.com/lxc/lxd/shared"
+	incus "github.com/lxc/incus/shared/util"
 	"github.com/sirupsen/logrus"
 	"github.com/spf13/cobra"
 	"gopkg.in/yaml.v2"
@@ -76,6 +77,9 @@ import (
 	"github.com/lxc/distrobuilder/sources"
 )
 
+//go:embed lxc.generator
+var lxcGenerator embed.FS
+
 var typeDescription = `Depending on the type, it either outputs a unified (single tarball)
 or split image (tarball + squashfs or qcow2 image). The --type flag can take one of the
 following values:
@@ -89,9 +93,13 @@ following values:
   - gzip
   - lzip
   - lzma
+  - lzo
   - lzop
   - xz (default)
   - zstd
+
+For supported compression methods, a compression level can be specified with
+method-N, where N is an integer, e.g. gzip-9.
 `
 
 type cmdGlobal struct {
@@ -113,6 +121,7 @@ type cmdGlobal struct {
 	overlayCleanup func()
 	ctx            context.Context
 	cancel         context.CancelFunc
+	subCommand     *cobra.Command
 }
 
 func main() {
@@ -121,7 +130,7 @@ func main() {
 
 	app := &cobra.Command{
 		Use:   "distrobuilder",
-		Short: "System container image builder for LXC and LXD",
+		Short: "System container and VM image builder for LXC and Incus",
 		PersistentPreRun: func(cmd *cobra.Command, args []string) {
 			// Quick checks
 			if os.Geteuid() != 0 {
@@ -129,6 +138,11 @@ func main() {
 				os.Exit(1)
 			}
 
+			// Keep track of the current subcommand. When app.Execute() fails, there's no way of
+			// knowing which command failed. However, we want to know this as we call postRun in
+			// case of an error, and handle the "validate" subcommand differently in that function.
+			globalCmd.subCommand = cmd
+
 			var err error
 
 			globalCmd.logger, err = shared.GetLogger(globalCmd.flagDebug)
@@ -154,14 +168,20 @@ func main() {
 						if globalCmd.flagTimeout > 0 {
 							globalCmd.logger.Info("Timed out")
 						}
+
 						return
 					}
 				}
 			}()
 
+			// No need to create cache directory if we're only validating.
+			if cmd.CalledAs() == "validate" {
+				return
+			}
+
 			// Create temp directory if the cache directory isn't explicitly set
 			if globalCmd.flagCacheDir == "" {
-				dir, err := ioutil.TempDir("/var/cache", "distrobuilder.")
+				dir, err := os.MkdirTemp("/var/cache", "distrobuilder.")
 				if err != nil {
 					fmt.Fprintf(os.Stderr, "Failed to create cache directory: %s\n", err)
 					os.Exit(1)
@@ -195,10 +215,10 @@ func main() {
 	app.AddCommand(LXCCmd.commandBuild())
 	app.AddCommand(LXCCmd.commandPack())
 
-	// LXD sub-commands
-	LXDCmd := cmdLXD{global: &globalCmd}
-	app.AddCommand(LXDCmd.commandBuild())
-	app.AddCommand(LXDCmd.commandPack())
+	// Incus sub-commands
+	IncusCmd := cmdIncus{global: &globalCmd}
+	app.AddCommand(IncusCmd.commandBuild())
+	app.AddCommand(IncusCmd.commandPack())
 
 	// build-dir sub-command
 	buildDirCmd := cmdBuildDir{global: &globalCmd}
@@ -217,11 +237,37 @@ func main() {
 	// Run the main command and handle errors
 	err := app.Execute()
 	if err != nil {
-		globalCmd.postRun(nil, nil)
+		if globalCmd.logger != nil {
+			globalCmd.logger.WithFields(logrus.Fields{"err": err}).Error("Failed running distrobuilder")
+		} else {
+			fmt.Fprintf(os.Stderr, "Failed running distrobuilder: %s\n", err.Error())
+		}
+
+		_ = globalCmd.postRun(globalCmd.subCommand, nil)
 		os.Exit(1)
 	}
 }
 
+func (c *cmdGlobal) cleanupCacheDirectory() {
+	// Try removing the entire cache directory.
+	err := os.RemoveAll(c.flagCacheDir)
+	if err == nil {
+		return
+	}
+
+	// Try removing the content of the cache directory if the directory itself cannot be removed.
+	err = filepath.Walk(c.flagCacheDir, func(path string, info os.FileInfo, err error) error {
+		if path == c.flagCacheDir {
+			return nil
+		}
+
+		return os.RemoveAll(path)
+	})
+	if err != nil {
+		c.logger.WithField("err", err).Warn("Failed cleaning up cache directory")
+	}
+}
+
 func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 	// if an error is returned, disable the usage message
 	cmd.SilenceUsage = true
@@ -229,8 +275,12 @@ func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 	isRunningBuildDir := cmd.CalledAs() == "build-dir"
 
 	// Clean up cache directory before doing anything
-	os.RemoveAll(c.flagCacheDir)
-	os.Mkdir(c.flagCacheDir, 0755)
+	c.cleanupCacheDirectory()
+
+	err := os.MkdirAll(c.flagCacheDir, 0755)
+	if err != nil {
+		return fmt.Errorf("Failed creating cache directory: %w", err)
+	}
 
 	if len(args) > 1 {
 		// Create and set target directory if provided
@@ -238,6 +288,7 @@ func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 		if err != nil {
 			return fmt.Errorf("Failed to create directory %q: %w", args[1], err)
 		}
+
 		c.targetDir = args[1]
 	} else {
 		// Use current working directory as target
@@ -254,7 +305,7 @@ func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 	}
 
 	// Create source directory if it doesn't exist
-	err := os.MkdirAll(c.sourceDir, 0755)
+	err = os.MkdirAll(c.sourceDir, 0755)
 	if err != nil {
 		return fmt.Errorf("Failed to create directory %q: %w", c.sourceDir, err)
 	}
@@ -265,7 +316,7 @@ func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 		return fmt.Errorf("Failed to get definition: %w", err)
 	}
 
-	// Create cache directory if we also plan on creating LXC or LXD images
+	// Create cache directory if we also plan on creating LXC or Incus images
 	if !isRunningBuildDir {
 		err = os.MkdirAll(c.flagCacheDir, 0755)
 		if err != nil {
@@ -301,18 +352,20 @@ func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 	}
 
 	// Setup the mounts and chroot into the rootfs
-	exitChroot, err := shared.SetupChroot(c.sourceDir, c.definition.Environment, nil)
+	exitChroot, err := shared.SetupChroot(c.sourceDir, *c.definition, nil)
 	if err != nil {
 		return fmt.Errorf("Failed to setup chroot: %w", err)
 	}
 	// Unmount everything and exit the chroot
-	defer exitChroot()
+	defer func() {
+		_ = exitChroot()
+	}()
 
 	// Always include sections which have no type filter. If running build-dir,
 	// only these sections will be processed.
 	imageTargets := shared.ImageTargetUndefined
 
-	// If we're running either build-lxc or build-lxd, include types which are
+	// If we're running either build-lxc or build-incus, include types which are
 	// meant for all.
 	if !isRunningBuildDir {
 		imageTargets |= shared.ImageTargetAll
@@ -322,9 +375,9 @@ func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 	case "build-lxc":
 		// If we're running build-lxc, also process container-only sections.
 		imageTargets |= shared.ImageTargetContainer
-	case "build-lxd":
+	case "build-incus":
 		// Include either container-specific or vm-specific sections when
-		// running build-lxd.
+		// running build-incus.
 		ok, err := cmd.Flags().GetBool("vm")
 		if err != nil {
 			return fmt.Errorf(`Failed to get bool value of "vm": %w`, err)
@@ -354,6 +407,13 @@ func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 
 	// Run post unpack hook
 	for _, hook := range c.definition.GetRunnableActions("post-unpack", imageTargets) {
+		if hook.Pongo {
+			hook.Action, err = shared.RenderTemplate(hook.Action, c.definition)
+			if err != nil {
+				return fmt.Errorf("Failed to render action: %w", err)
+			}
+		}
+
 		err := shared.RunScript(c.ctx, hook.Action)
 		if err != nil {
 			return fmt.Errorf("Failed to run post-unpack: %w", err)
@@ -372,6 +432,13 @@ func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 
 	// Run post packages hook
 	for _, hook := range c.definition.GetRunnableActions("post-packages", imageTargets) {
+		if hook.Pongo {
+			hook.Action, err = shared.RenderTemplate(hook.Action, c.definition)
+			if err != nil {
+				return fmt.Errorf("Failed to render action: %w", err)
+			}
+		}
+
 		err := shared.RunScript(c.ctx, hook.Action)
 		if err != nil {
 			return fmt.Errorf("Failed to run post-packages: %w", err)
@@ -382,14 +449,16 @@ func (c *cmdGlobal) preRunBuild(cmd *cobra.Command, args []string) error {
 }
 
 func (c *cmdGlobal) preRunPack(cmd *cobra.Command, args []string) error {
-	var err error
-
 	// if an error is returned, disable the usage message
 	cmd.SilenceUsage = true
 
 	// Clean up cache directory before doing anything
-	os.RemoveAll(c.flagCacheDir)
-	os.Mkdir(c.flagCacheDir, 0755)
+	c.cleanupCacheDirectory()
+
+	err := os.MkdirAll(c.flagCacheDir, 0755)
+	if err != nil {
+		return fmt.Errorf("Failed creating cache directory: %w", err)
+	}
 
 	// resolve path
 	c.sourceDir, err = filepath.Abs(args[1])
@@ -412,12 +481,20 @@ func (c *cmdGlobal) preRunPack(cmd *cobra.Command, args []string) error {
 }
 
 func (c *cmdGlobal) postRun(cmd *cobra.Command, args []string) error {
+	// If we're only validating, there's nothing to clean up.
+	if cmd != nil && cmd.CalledAs() == "validate" {
+		return nil
+	}
+
 	hasLogger := c.logger != nil
 
 	// exit all chroots otherwise we cannot remove the cache directory
 	for _, exit := range shared.ActiveChroots {
 		if exit != nil {
-			exit()
+			err := exit()
+			if err != nil && hasLogger {
+				c.logger.WithField("err", err).Warn("Failed exiting chroot")
+			}
 		}
 	}
 
@@ -436,7 +513,7 @@ func (c *cmdGlobal) postRun(cmd *cobra.Command, args []string) error {
 			c.logger.Info("Removing cache directory")
 		}
 
-		os.RemoveAll(c.flagCacheDir)
+		c.cleanupCacheDirectory()
 	}
 
 	// Clean up sources directory
@@ -445,7 +522,7 @@ func (c *cmdGlobal) postRun(cmd *cobra.Command, args []string) error {
 			c.logger.Info("Removing sources directory")
 		}
 
-		os.RemoveAll(c.flagSourcesDir)
+		_ = os.RemoveAll(c.flagSourcesDir)
 	}
 
 	return nil
@@ -497,6 +574,7 @@ func getDefinition(fname string, options []string) (*shared.Definition, error) {
 		if err != nil {
 			return nil, err
 		}
+
 		defer f.Close()
 
 		_, err = io.Copy(&buf, f)
@@ -538,251 +616,26 @@ func getDefinition(fname string, options []string) (*shared.Definition, error) {
 }
 
 // addSystemdGenerator creates a systemd-generator which runs on boot, and does some configuration around the system itself and networking.
-func addSystemdGenerator() {
+func addSystemdGenerator() error {
 	// Check if container has systemd
-	if !lxd.PathExists("/etc/systemd") {
-		return
+	if !incus.PathExists("/etc/systemd") {
+		return nil
 	}
 
-	content := `#!/bin/sh
-# NOTE: systemctl is not available for systemd-generators
-set -eu
+	err := os.MkdirAll("/etc/systemd/system-generators", 0755)
+	if err != nil {
+		return fmt.Errorf("Failed creating directory: %w", err)
+	}
 
-## Helper functions
-# is_lxc_container succeeds if we're running inside a LXC container
-is_lxc_container() {
-	grep -qa container=lxc /proc/1/environ
-}
+	content, err := lxcGenerator.ReadFile("lxc.generator")
+	if err != nil {
+		return fmt.Errorf("Failed reading lxc.generator: %w", err)
+	}
 
-# is_lxd_vm succeeds if we're running inside a LXD VM
-is_lxd_vm() {
-	[ -e /dev/virtio-ports/org.linuxcontainers.lxd ]
-}
+	err = os.WriteFile("/etc/systemd/system-generators/lxc", content, 0755)
+	if err != nil {
+		return fmt.Errorf("Failed creating system generator: %w", err)
+	}
 
-# is_in_path succeeds if the given file exists in on of the paths
-is_in_path() {
-	# Don't use $PATH as that may not include all relevant paths
-	for path in /bin /sbin /usr/bin /usr/sbin /usr/local/bin /usr/local/sbin; do
-		[ -e "${path}/$1" ] && return 0
-	done
-
-	return 1
-}
-
-## Fix functions
-# fix_ro_paths avoids udevd issues with /sys and /proc being writable
-fix_ro_paths() {
-	mkdir -p "/run/systemd/system/$1.d"
-	cat <<-EOF > "/run/systemd/system/$1.d/zzz-lxc-ropath.conf"
-[Service]
-BindReadOnlyPaths=/sys /proc
-EOF
-}
-
-# fix_nm_force_up sets up a unit override to force NetworkManager to start the system connection
-fix_nm_force_up() {
-	# Check if the device exists
-	[ -e "/sys/class/net/$1" ] || return 0
-
-	# Check if NetworkManager exists
-	[ "${nm_exists}" -eq 1 ] || return 0
-
-	cat <<-EOF > /run/systemd/system/network-connection-activate.service
-[Unit]
-Description=Activate connection
-After=NetworkManager.service NetworkManager-wait-online.service
-
-[Service]
-ExecStart=-/usr/bin/nmcli c up "System $1"
-Type=oneshot
-RemainAfterExit=true
-
-[Install]
-WantedBy=default.target
-EOF
-
-	mkdir -p /run/systemd/system/default.target.wants
-	ln -sf /run/systemd/system/network-connection-activate.service /run/systemd/system/default.target.wants/network-connection-activate.service
-}
-
-# fix_nm_link_state forces the network interface to a DOWN state ahead of NetworkManager starting up
-fix_nm_link_state() {
-	[ -e "/sys/class/net/$1" ] || return 0
-
-	ip_path=
-	if [ -f /sbin/ip ]; then
-		ip_path=/sbin/ip
-	elif [ -f /bin/ip ]; then
-		ip_path=/bin/ip
-	else
-		return 0
-	fi
-
-	cat <<-EOF > /run/systemd/system/network-device-down.service
-[Unit]
-Description=Turn off network device
-Before=NetworkManager.service
-Before=systemd-networkd.service
-
-[Service]
-ExecStart=-${ip_path} link set $1 down
-Type=oneshot
-RemainAfterExit=true
-
-[Install]
-WantedBy=default.target
-EOF
-
-	mkdir -p /run/systemd/system/default.target.wants
-	ln -sf /run/systemd/system/network-device-down.service /run/systemd/system/default.target.wants/network-device-down.service
-}
-
-# fix_systemd_override_unit generates a unit specific override
-fix_systemd_override_unit() {
-	dropin_dir="/run/systemd/${1}.d"
-	mkdir -p "${dropin_dir}"
-	{
-		echo "[Service]";
-		[ "${systemd_version}" -ge 247 ] && echo "ProcSubset=all";
-		[ "${systemd_version}" -ge 247 ] && echo "ProtectProc=default";
-		[ "${systemd_version}" -ge 232 ] && echo "ProtectControlGroups=no";
-		[ "${systemd_version}" -ge 232 ] && echo "ProtectKernelTunables=no";
-		[ "${systemd_version}" -ge 239 ] && echo "NoNewPrivileges=no";
-		[ "${systemd_version}" -ge 249 ] && echo "LoadCredential=";
-
-		# Additional settings for privileged containers
-		if grep -q 4294967295 /proc/self/uid_map; then
-			echo "ProtectHome=no";
-			echo "ProtectSystem=no";
-			echo "PrivateDevices=no";
-			echo "PrivateTmp=no";
-			[ "${systemd_version}" -ge 244 ] && echo "ProtectKernelLogs=no";
-			[ "${systemd_version}" -ge 232 ] && echo "ProtectKernelModules=no";
-			[ "${systemd_version}" -ge 231 ] && echo "ReadWritePaths=";
-		fi
-	} > "${dropin_dir}/zzz-lxc-service.conf"
-}
-
-# fix_systemd_mask masks the systemd unit
-fix_systemd_mask() {
-	ln -sf /dev/null "/run/systemd/system/$1"
-}
-
-# fix_systemd_udev_trigger overrides the systemd-udev-trigger.service to match the latest version
-# of the file which uses "ExecStart=-" instead of "ExecStart=".
-fix_systemd_udev_trigger() {
-	cmd=
-	if [ -f /usr/bin/udevadm ]; then
-		cmd=/usr/bin/udevadm
-	elif [ -f /sbin/udevadm ]; then
-		cmd=/sbin/udevadm
-	elif [ -f /bin/udevadm ]; then
-		cmd=/bin/udevadm
-	else
-		return 0
-	fi
-
-	mkdir -p /run/systemd/system/systemd-udev-trigger.service.d
-	cat <<-EOF > /run/systemd/system/systemd-udev-trigger.service.d/zzz-lxc-override.conf
-[Service]
-ExecStart=
-ExecStart=-${cmd} trigger --type=subsystems --action=add
-ExecStart=-${cmd} trigger --type=devices --action=add
-EOF
-}
-
-# fix_systemd_sysctl overrides the systemd-sysctl.service to use "ExecStart=-" instead of "ExecStart=".
-fix_systemd_sysctl() {
-	cmd=/usr/lib/systemd/systemd-sysctl
-	! [ -e "${cmd}" ] && cmd=/lib/systemd/systemd-sysctl
-	mkdir -p /run/systemd/system/systemd-sysctl.service.d
-	cat <<-EOF > /run/systemd/system/systemd-sysctl.service.d/zzz-lxc-override.conf
-[Service]
-ExecStart=
-ExecStart=-${cmd}
-EOF
-}
-
-## Main logic
-# Exit immediately if not a LXC/LXD container or VM
-if ! is_lxd_vm && ! is_lxc_container; then
-	exit
-fi
-
-# Check for NetworkManager and cloud-init
-nm_exists=0
-cloudinit_exists=0
-
-is_in_path NetworkManager && nm_exists=1
-is_in_path cloud-init && cloudinit_exists=1
-
-# Determine systemd version
-for path in /usr/lib/systemd/systemd /lib/systemd/systemd; do
-	[ -x "${path}" ] || continue
-
-	systemd_version="$("${path}" --version | head -n1 | cut -d' ' -f2)"
-	break
-done
-
-# Determine distro name and release
-ID=""
-if [ -e /etc/os-release ]; then
-	. /etc/os-release
-fi
-
-# Workarounds for containers.
-if is_lxc_container; then
-	# Overriding some systemd features is only needed if security.nesting=false
-	# in which case, /dev/.lxc will be missing
-	if [ ! -d /dev/.lxc ]; then
-		# Apply systemd overrides
-		if [ "${systemd_version}" -ge 244 ]; then
-			fix_systemd_override_unit system/service
-		else
-			# Setup per-unit overrides
-			find /lib/systemd /etc/systemd /run/systemd /usr/lib/systemd -name "*.service" -type f | sed 's#/\(lib\|etc\|run\|usr/lib\)/systemd/##g'| while read -r service_file; do
-				fix_systemd_override_unit "${service_file}"
-			done
-		fi
-
-		# Workarounds for privileged containers.
-		if ! grep -q 4294967295 /proc/self/uid_map && { [ "${ID}" = "altlinux" ] || [ "${ID}" = "arch" ] || [ "${ID}" = "fedora" ]; }; then
-			fix_ro_paths systemd-networkd.service
-			fix_ro_paths systemd-resolved.service
-		fi
-	fi
-
-	# Ignore failures on some units.
-	fix_systemd_udev_trigger
-	fix_systemd_sysctl
-
-	# Mask some units.
-	fix_systemd_mask dev-hugepages.mount
-	fix_systemd_mask run-ribchester-general.mount
-	fix_systemd_mask systemd-hwdb-update.service
-	fix_systemd_mask systemd-journald-audit.socket
-	fix_systemd_mask systemd-modules-load.service
-	fix_systemd_mask systemd-pstore.service
-	fix_systemd_mask ua-messaging.service
-	if [ ! -e /dev/tty1 ]; then
-		fix_systemd_mask vconsole-setup-kludge@tty1.service
-	fi
-
-	# Workarounds for cloud containers
-	if { [ "${ID}" = "fedora" ] || [ "${ID}" = "rhel" ]; } && [ "${cloudinit_exists}" -eq 1 ]; then
-		fix_nm_force_up eth0
-	fi
-
-	# Workarounds for NetworkManager in containers
-	if [ "${nm_exists}" -eq 1 ]; then
-		if [ "${ID}" = "ol" ] || [ "${ID}" = "centos" ]; then
-			fix_nm_force_up eth0
-		fi
-
-		fix_nm_link_state eth0
-	fi
-fi
-`
-	os.MkdirAll("/etc/systemd/system-generators", 0755)
-	ioutil.WriteFile("/etc/systemd/system-generators/lxc", []byte(content), 0755)
+	return nil
 }
