@@ -6,7 +6,6 @@ import (
 	"encoding/hex"
 	"errors"
 	"fmt"
-	"io/ioutil"
 	"net/http"
 	"os"
 	"os/exec"
@@ -16,7 +15,7 @@ import (
 	"strings"
 
 	"github.com/flosch/pongo2"
-	lxd "github.com/lxc/lxd/shared"
+	incus "github.com/lxc/incus/shared/util"
 	"github.com/sirupsen/logrus"
 	"github.com/spf13/cobra"
 	"golang.org/x/sys/unix"
@@ -28,13 +27,14 @@ import (
 type cmdRepackWindows struct {
 	global *cmdGlobal
 
-	flagDrivers        string
-	flagWindowsVersion string
+	flagDrivers             string
+	flagWindowsVersion      string
+	flagWindowsArchitecture string
 }
 
 func init() {
 	// Filters should be registered in the init() function
-	pongo2.RegisterFilter("toHex", toHex)
+	_ = pongo2.RegisterFilter("toHex", toHex)
 }
 
 func (c *cmdRepackWindows) command() *cobra.Command {
@@ -44,7 +44,9 @@ func (c *cmdRepackWindows) command() *cobra.Command {
 		Args:    cobra.ExactArgs(2),
 		PreRunE: c.preRun,
 		RunE: func(cmd *cobra.Command, args []string) error {
-			defer unix.Unmount(c.global.sourceDir, 0)
+			defer func() {
+				_ = unix.Unmount(c.global.sourceDir, 0)
+			}()
 
 			sourceDir := filepath.Dir(args[0])
 			targetDir := filepath.Dir(args[1])
@@ -88,6 +90,7 @@ func (c *cmdRepackWindows) command() *cobra.Command {
 
 	cmd.Flags().StringVar(&c.flagDrivers, "drivers", "", "Path to drivers ISO"+"``")
 	cmd.Flags().StringVar(&c.flagWindowsVersion, "windows-version", "", "Windows version to repack"+"``")
+	cmd.Flags().StringVar(&c.flagWindowsArchitecture, "windows-arch", "", "Windows architecture to repack"+"``")
 
 	return cmd
 }
@@ -107,11 +110,27 @@ func (c *cmdRepackWindows) preRun(cmd *cobra.Command, args []string) error {
 	} else {
 		supportedVersions := []string{"w11", "w10", "2k19", "2k12", "2k16", "2k22"}
 
-		if !lxd.StringInSlice(c.flagWindowsVersion, supportedVersions) {
+		if !incus.ValueInSlice(c.flagWindowsVersion, supportedVersions) {
 			return fmt.Errorf("Version must be one of %v", supportedVersions)
 		}
 	}
 
+	if c.flagWindowsArchitecture == "" {
+		detectedArchitecture := detectWindowsArchitecture(filepath.Base(args[0]))
+
+		if detectedArchitecture == "" {
+			return errors.New("Failed to detect Windows architecture. Please provide the architecture using the --windows-arch flag")
+		}
+
+		c.flagWindowsArchitecture = detectedArchitecture
+	} else {
+		supportedArchitectures := []string{"amd64", "ARM64"}
+
+		if !incus.ValueInSlice(c.flagWindowsArchitecture, supportedArchitectures) {
+			return fmt.Errorf("Architecture must be one of %v", supportedArchitectures)
+		}
+	}
+
 	// Check dependencies
 	err := c.checkDependencies()
 	if err != nil {
@@ -133,6 +152,7 @@ func (c *cmdRepackWindows) preRun(cmd *cobra.Command, args []string) error {
 	if err != nil {
 		return fmt.Errorf("Failed to create directory %q: %w", c.global.flagCacheDir, err)
 	}
+
 	defer func() {
 		if c.global.flagCleanup && !success {
 			os.RemoveAll(c.global.flagCacheDir)
@@ -150,7 +170,7 @@ func (c *cmdRepackWindows) preRun(cmd *cobra.Command, args []string) error {
 	logger.Info("Mounting Windows ISO")
 
 	// Mount ISO
-	err = shared.RunCommand(c.global.ctx, nil, nil, "mount", "-o", "loop", args[0], c.global.sourceDir)
+	err = shared.RunCommand(c.global.ctx, nil, nil, "mount", "-t", "udf", "-o", "loop", args[0], c.global.sourceDir)
 	if err != nil {
 		return fmt.Errorf("Failed to mount %q at %q: %w", args[0], c.global.sourceDir, err)
 	}
@@ -171,7 +191,7 @@ func (c *cmdRepackWindows) run(cmd *cobra.Command, args []string, overlayDir str
 
 		virtioISOPath = filepath.Join(c.global.flagSourcesDir, "windows", "virtio-win.iso")
 
-		if !lxd.PathExists(virtioISOPath) {
+		if !incus.PathExists(virtioISOPath) {
 			err := os.MkdirAll(filepath.Dir(virtioISOPath), 0755)
 			if err != nil {
 				return fmt.Errorf("Failed to create directory %q: %w", filepath.Dir(virtioISOPath), err)
@@ -181,13 +201,14 @@ func (c *cmdRepackWindows) run(cmd *cobra.Command, args []string, overlayDir str
 			if err != nil {
 				return fmt.Errorf("Failed to create file %q: %w", virtioISOPath, err)
 			}
+
 			defer f.Close()
 
 			var client http.Client
 
 			logger.Info("Downloading drivers ISO")
 
-			_, err = lxd.DownloadFileHash(c.global.ctx, &client, "", nil, nil, "virtio-win.iso", virtioURL, "", nil, f)
+			_, err = incus.DownloadFileHash(c.global.ctx, &client, "", nil, nil, "virtio-win.iso", virtioURL, "", nil, f)
 			if err != nil {
 				f.Close()
 				os.Remove(virtioISOPath)
@@ -198,7 +219,7 @@ func (c *cmdRepackWindows) run(cmd *cobra.Command, args []string, overlayDir str
 		}
 	}
 
-	if !lxd.PathExists(driverPath) {
+	if !incus.PathExists(driverPath) {
 		err := os.MkdirAll(driverPath, 0755)
 		if err != nil {
 			return fmt.Errorf("Failed to create directory %q: %w", driverPath, err)
@@ -208,15 +229,18 @@ func (c *cmdRepackWindows) run(cmd *cobra.Command, args []string, overlayDir str
 	logger.Info("Mounting driver ISO")
 
 	// Mount driver ISO
-	err := shared.RunCommand(c.global.ctx, nil, nil, "mount", "-o", "loop", virtioISOPath, driverPath)
+	err := shared.RunCommand(c.global.ctx, nil, nil, "mount", "-t", "iso9660", "-o", "loop", virtioISOPath, driverPath)
 	if err != nil {
 		return fmt.Errorf("Failed to mount %q at %q: %w", virtioISOPath, driverPath, err)
 	}
-	defer unix.Unmount(driverPath, 0)
+
+	defer func() {
+		_ = unix.Unmount(driverPath, 0)
+	}()
 
 	var sourcesDir string
 
-	entries, err := ioutil.ReadDir(overlayDir)
+	entries, err := os.ReadDir(overlayDir)
 	if err != nil {
 		return fmt.Errorf("Failed to read directory %q: %w", overlayDir, err)
 	}
@@ -228,7 +252,7 @@ func (c *cmdRepackWindows) run(cmd *cobra.Command, args []string, overlayDir str
 		}
 	}
 
-	entries, err = ioutil.ReadDir(sourcesDir)
+	entries, err = os.ReadDir(sourcesDir)
 	if err != nil {
 		return fmt.Errorf("Failed to read directory %q: %w", sourcesDir, err)
 	}
@@ -281,6 +305,7 @@ func (c *cmdRepackWindows) run(cmd *cobra.Command, args []string, overlayDir str
 			if err != nil {
 				return fmt.Errorf("Failed to determine wim file indexes: %w", err)
 			}
+
 			indexes = append(indexes, index)
 		}
 	}
@@ -314,6 +339,7 @@ func (c *cmdRepackWindows) run(cmd *cobra.Command, args []string, overlayDir str
 	} else {
 		err = shared.RunCommand(c.global.ctx, nil, nil, "genisoimage", "--allow-limited-size", "-l", "-no-emul-boot", "-b", "efi/microsoft/boot/efisys.bin", "-o", args[1], overlayDir)
 	}
+
 	if err != nil {
 		return fmt.Errorf("Failed to generate ISO: %w", err)
 	}
@@ -328,7 +354,7 @@ func (c *cmdRepackWindows) modifyWim(path string, index int) error {
 	wimFile := filepath.Join(path)
 	wimPath := filepath.Join(c.global.flagCacheDir, "wim")
 
-	if !lxd.PathExists(wimPath) {
+	if !incus.PathExists(wimPath) {
 		err := os.MkdirAll(wimPath, 0755)
 		if err != nil {
 			return fmt.Errorf("Failed to create directory %q: %w", wimPath, err)
@@ -341,9 +367,10 @@ func (c *cmdRepackWindows) modifyWim(path string, index int) error {
 	if err != nil {
 		return fmt.Errorf("Failed to mount %q: %w", filepath.Base(wimFile), err)
 	}
+
 	defer func() {
 		if !success {
-			shared.RunCommand(c.global.ctx, nil, nil, "wimlib-imagex", "unmount", wimPath)
+			_ = shared.RunCommand(c.global.ctx, nil, nil, "wimlib-imagex", "unmount", wimPath)
 		}
 	}()
 
@@ -404,7 +431,7 @@ func (c *cmdRepackWindows) getWindowsDirectories(wimPath string) (map[string]str
 	driverStorePath := ""
 	dirs := make(map[string]string)
 
-	entries, err := ioutil.ReadDir(wimPath)
+	entries, err := os.ReadDir(wimPath)
 	if err != nil {
 		return nil, err
 	}
@@ -421,7 +448,7 @@ func (c *cmdRepackWindows) getWindowsDirectories(wimPath string) (map[string]str
 		}
 	}
 
-	entries, err = ioutil.ReadDir(windowsPath)
+	entries, err = os.ReadDir(windowsPath)
 	if err != nil {
 		return nil, err
 	}
@@ -445,7 +472,7 @@ func (c *cmdRepackWindows) getWindowsDirectories(wimPath string) (map[string]str
 		}
 	}
 
-	entries, err = ioutil.ReadDir(system32Path)
+	entries, err = os.ReadDir(system32Path)
 	if err != nil {
 		return nil, err
 	}
@@ -474,7 +501,7 @@ func (c *cmdRepackWindows) getWindowsDirectories(wimPath string) (map[string]str
 		}
 	}
 
-	entries, err = ioutil.ReadDir(driverStorePath)
+	entries, err = os.ReadDir(driverStorePath)
 	if err != nil {
 		return nil, err
 	}
@@ -512,10 +539,10 @@ func (c *cmdRepackWindows) injectDrivers(dirs map[string]string) error {
 			"driverName":  driver,
 		}
 
-		sourceDir := filepath.Join(driverPath, driver, c.flagWindowsVersion, "amd64")
+		sourceDir := filepath.Join(driverPath, driver, c.flagWindowsVersion, c.flagWindowsArchitecture)
 		targetBasePath := filepath.Join(dirs["filerepository"], info.PackageName)
 
-		if !lxd.PathExists(targetBasePath) {
+		if !incus.PathExists(targetBasePath) {
 			err := os.MkdirAll(targetBasePath, 0755)
 			if err != nil {
 				return fmt.Errorf("Failed to create directory %q: %w", targetBasePath, err)
@@ -527,7 +554,7 @@ func (c *cmdRepackWindows) injectDrivers(dirs map[string]string) error {
 			targetPath := filepath.Join(targetBasePath, filepath.Base(path))
 
 			// Copy driver files
-			if lxd.StringInSlice(ext, []string{".cat", ".dll", ".inf", ".sys"}) {
+			if incus.ValueInSlice(ext, []string{".cat", ".dll", ".inf", ".sys"}) {
 				logger.WithFields(logrus.Fields{"src": path, "dest": targetPath}).Debug("Copying file")
 
 				err := shared.Copy(path, targetPath)
@@ -681,6 +708,23 @@ func detectWindowsVersion(fileName string) string {
 	return ""
 }
 
+func detectWindowsArchitecture(fileName string) string {
+	aliases := map[string][]string{
+		"amd64": {"amd64", "x64"},
+		"ARM64": {"arm64"},
+	}
+
+	for k, v := range aliases {
+		for _, alias := range v {
+			if regexp.MustCompile(fmt.Sprintf("(?i)%s", alias)).MatchString(fileName) {
+				return k
+			}
+		}
+	}
+
+	return ""
+}
+
 // toHex is a pongo2 filter which converts the provided value to a hex value understood by the Windows registry.
 func toHex(in *pongo2.Value, param *pongo2.Value) (out *pongo2.Value, err *pongo2.Error) {
 	dst := make([]byte, hex.EncodedLen(len(in.String())))
