@@ -9,7 +9,7 @@ import (
 	"strconv"
 	"strings"
 
-	lxd "github.com/lxc/lxd/shared"
+	incus "github.com/lxc/incus/shared/util"
 	"golang.org/x/sys/unix"
 
 	"github.com/lxc/distrobuilder/shared"
@@ -29,7 +29,7 @@ func newVM(ctx context.Context, imageFile, rootfsDir, fs string, size uint64) (*
 		fs = "ext4"
 	}
 
-	if !lxd.StringInSlice(fs, []string{"btrfs", "ext4"}) {
+	if !incus.ValueInSlice(fs, []string{"btrfs", "ext4"}) {
 		return nil, fmt.Errorf("Unsupported fs: %s", fs)
 	}
 
@@ -65,6 +65,7 @@ func (v *vm) createEmptyDiskImage() error {
 	if err != nil {
 		return fmt.Errorf("Failed to open %s: %w", v.imageFile, err)
 	}
+
 	defer f.Close()
 
 	err = f.Chmod(0600)
@@ -123,7 +124,7 @@ func (v *vm) mountImage() error {
 
 	deviceNumbers := strings.Split(out.String(), "\n")
 
-	if !lxd.PathExists(v.getUEFIDevFile()) {
+	if !incus.PathExists(v.getUEFIDevFile()) {
 		fields := strings.Split(deviceNumbers[1], ":")
 
 		major, err := strconv.Atoi(fields[0])
@@ -144,7 +145,7 @@ func (v *vm) mountImage() error {
 		}
 	}
 
-	if !lxd.PathExists(v.getRootfsDevFile()) {
+	if !incus.PathExists(v.getRootfsDevFile()) {
 		fields := strings.Split(deviceNumbers[2], ":")
 
 		major, err := strconv.Atoi(fields[0])
@@ -170,7 +171,7 @@ func (v *vm) mountImage() error {
 
 func (v *vm) umountImage() error {
 	// If loopDevice is empty, the image probably isn't mounted.
-	if v.loopDevice == "" || !lxd.PathExists(v.loopDevice) {
+	if v.loopDevice == "" || !incus.PathExists(v.loopDevice) {
 		return nil
 	}
 
@@ -180,14 +181,14 @@ func (v *vm) umountImage() error {
 	}
 
 	// Make sure that p1 and p2 are also removed.
-	if lxd.PathExists(v.getUEFIDevFile()) {
+	if incus.PathExists(v.getUEFIDevFile()) {
 		err := os.Remove(v.getUEFIDevFile())
 		if err != nil {
 			return fmt.Errorf("Failed to remove file %q: %w", v.getUEFIDevFile(), err)
 		}
 	}
 
-	if lxd.PathExists(v.getRootfsDevFile()) {
+	if incus.PathExists(v.getRootfsDevFile()) {
 		err := os.Remove(v.getRootfsDevFile())
 		if err != nil {
 			return fmt.Errorf("Failed to remove file %q: %w", v.getRootfsDevFile(), err)
@@ -212,11 +213,15 @@ func (v *vm) createRootFS() error {
 		}
 
 		// Create the root subvolume as well
-		err = shared.RunCommand(v.ctx, nil, nil, "mount", v.getRootfsDevFile(), v.rootfsDir)
+
+		err = shared.RunCommand(v.ctx, nil, nil, "mount", "-t", v.rootFS, v.getRootfsDevFile(), v.rootfsDir)
 		if err != nil {
 			return fmt.Errorf("Failed to mount %q at %q: %w", v.getRootfsDevFile(), v.rootfsDir, err)
 		}
-		defer shared.RunCommand(v.ctx, nil, nil, "umount", v.rootfsDir)
+
+		defer func() {
+			_ = shared.RunCommand(v.ctx, nil, nil, "umount", v.rootfsDir)
+		}()
 
 		return shared.RunCommand(v.ctx, nil, nil, "btrfs", "subvolume", "create", fmt.Sprintf("%s/@", v.rootfsDir))
 	case "ext4":
@@ -241,10 +246,9 @@ func (v *vm) mountRootPartition() error {
 
 	switch v.rootFS {
 	case "btrfs":
-		return shared.RunCommand(v.ctx, nil, nil, "mount", v.getRootfsDevFile(), v.rootfsDir, "-o", "defaults,discard,nobarrier,commit=300,noatime,subvol=/@")
+		return shared.RunCommand(v.ctx, nil, nil, "mount", v.getRootfsDevFile(), v.rootfsDir, "-t", v.rootFS, "-o", "defaults,discard,nobarrier,commit=300,noatime,subvol=/@")
 	case "ext4":
-		return shared.RunCommand(v.ctx, nil, nil, "mount", v.getRootfsDevFile(), v.rootfsDir, "-o", "discard,nobarrier,commit=300,noatime,data=writeback")
-
+		return shared.RunCommand(v.ctx, nil, nil, "mount", v.getRootfsDevFile(), v.rootfsDir, "-t", v.rootFS, "-o", "discard,nobarrier,commit=300,noatime,data=writeback")
 	}
 
 	return nil
@@ -262,5 +266,5 @@ func (v *vm) mountUEFIPartition() error {
 		return fmt.Errorf("Failed to create directory %q: %w", mountpoint, err)
 	}
 
-	return shared.RunCommand(v.ctx, nil, nil, "mount", v.getUEFIDevFile(), mountpoint, "-o", "discard")
+	return shared.RunCommand(v.ctx, nil, nil, "mount", "-t", "vfat", v.getUEFIDevFile(), mountpoint, "-o", "discard")
 }
