@@ -8,8 +8,9 @@ import (
 	"strings"
 	"time"
 
-	"github.com/lxc/lxd/shared"
-	lxdarch "github.com/lxc/lxd/shared/osarch"
+	"github.com/lxc/incus/shared/osarch"
+	incusArch "github.com/lxc/incus/shared/osarch"
+	"github.com/lxc/incus/shared/util"
 )
 
 // ImageTarget represents the image target.
@@ -104,7 +105,7 @@ type DefinitionPackagesSet struct {
 	Flags            []string `yaml:"flags,omitempty"`
 }
 
-// A DefinitionPackagesRepository contains data of a specific repository
+// A DefinitionPackagesRepository contains data of a specific repository.
 type DefinitionPackagesRepository struct {
 	DefinitionFilter `yaml:",inline"`
 	Name             string `yaml:"name"`           // Name of the repository
@@ -156,7 +157,7 @@ type DefinitionImage struct {
 	ArchitecturePersonality string `yaml:"architecture_personality,omitempty"`
 }
 
-// A DefinitionSource specifies the download type and location
+// A DefinitionSource specifies the download type and location.
 type DefinitionSource struct {
 	Downloader       string   `yaml:"downloader"`
 	URL              string   `yaml:"url,omitempty"`
@@ -166,14 +167,16 @@ type DefinitionSource struct {
 	Suite            string   `yaml:"suite,omitempty"`
 	SameAs           string   `yaml:"same_as,omitempty"`
 	SkipVerification bool     `yaml:"skip_verification,omitempty"`
+	Components       []string `yaml:"components,omitempty"`
 }
 
 // A DefinitionTargetLXCConfig represents the config part of the metadata.
 type DefinitionTargetLXCConfig struct {
-	Type    string `yaml:"type"`
-	Before  uint   `yaml:"before,omitempty"`
-	After   uint   `yaml:"after,omitempty"`
-	Content string `yaml:"content"`
+	DefinitionFilter `yaml:",inline"`
+	Type             string `yaml:"type"`
+	Before           uint   `yaml:"before,omitempty"`
+	After            uint   `yaml:"after,omitempty"`
+	Content          string `yaml:"content"`
 }
 
 // A DefinitionTargetLXC represents LXC specific files as part of the metadata.
@@ -182,22 +185,22 @@ type DefinitionTargetLXC struct {
 	Config        []DefinitionTargetLXCConfig `yaml:"config,omitempty"`
 }
 
-// DefinitionTargetLXDVM represents LXD VM specific options.
-type DefinitionTargetLXDVM struct {
+// DefinitionTargetIncusVM represents Incus VM specific options.
+type DefinitionTargetIncusVM struct {
 	Size       uint64 `yaml:"size,omitempty"`
 	Filesystem string `yaml:"filesystem,omitempty"`
 }
 
-// DefinitionTargetLXD represents LXD specific options.
-type DefinitionTargetLXD struct {
-	VM DefinitionTargetLXDVM `yaml:"vm,omitempty"`
+// DefinitionTargetIncus represents Incus specific options.
+type DefinitionTargetIncus struct {
+	VM DefinitionTargetIncusVM `yaml:"vm,omitempty"`
 }
 
 // A DefinitionTarget specifies target dependent files.
 type DefinitionTarget struct {
-	LXC  DefinitionTargetLXC  `yaml:"lxc,omitempty"`
-	LXD  DefinitionTargetLXD  `yaml:"lxd,omitempty"`
-	Type DefinitionFilterType // This field is internal only and used only for simplicity.
+	LXC   DefinitionTargetLXC   `yaml:"lxc,omitempty"`
+	Incus DefinitionTargetIncus `yaml:"incus,omitempty"`
+	Type  DefinitionFilterType  // This field is internal only and used only for simplicity.
 }
 
 // A DefinitionFile represents a file which is to be created inside to chroot.
@@ -216,7 +219,7 @@ type DefinitionFile struct {
 	Source           string                 `yaml:"source,omitempty"`
 }
 
-// A DefinitionFileTemplate represents the settings used by generators
+// A DefinitionFileTemplate represents the settings used by generators.
 type DefinitionFileTemplate struct {
 	Properties map[string]string `yaml:"properties,omitempty"`
 	When       []string          `yaml:"when,omitempty"`
@@ -228,6 +231,7 @@ type DefinitionAction struct {
 	DefinitionFilter `yaml:",inline"`
 	Trigger          string `yaml:"trigger"`
 	Action           string `yaml:"action"`
+	Pongo            bool   `yaml:"pongo,omitempty"`
 }
 
 // DefinitionMappings defines custom mappings.
@@ -238,8 +242,9 @@ type DefinitionMappings struct {
 
 // DefinitionEnvVars defines custom environment variables.
 type DefinitionEnvVars struct {
-	Key   string `yaml:"key"`
-	Value string `yaml:"value"`
+	DefinitionFilter `yaml:",inline"`
+	Key              string `yaml:"key"`
+	Value            string `yaml:"value"`
 }
 
 // DefinitionEnv represents the config part of the environment section.
@@ -279,12 +284,14 @@ func (d *Definition) SetValue(key string, value string) error {
 		if err != nil {
 			return fmt.Errorf("Failed to parse bool %q: %w", value, err)
 		}
+
 		field.SetBool(v)
 	case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
 		v, err := strconv.ParseInt(value, 10, 64)
 		if err != nil {
 			return fmt.Errorf("Failed to parse int %q: %w", value, err)
 		}
+
 		field.SetInt(v)
 	case reflect.String:
 		field.SetString(value)
@@ -293,6 +300,7 @@ func (d *Definition) SetValue(key string, value string) error {
 		if err != nil {
 			return fmt.Errorf("Failed to parse uint %q: %w", value, err)
 		}
+
 		field.SetUint(v)
 	default:
 		return fmt.Errorf("Unsupported type '%s'", field.Kind())
@@ -305,8 +313,8 @@ func (d *Definition) SetValue(key string, value string) error {
 func (d *Definition) SetDefaults() {
 	// default to local arch
 	if d.Image.Architecture == "" {
-		uname, _ := shared.Uname()
-		d.Image.Architecture = uname.Machine
+		localArch, _ := osarch.ArchitectureGetLocal()
+		d.Image.Architecture = localArch
 	}
 
 	// set default expiry of 30 days
@@ -338,7 +346,7 @@ func (d *Definition) SetDefaults() {
 		d.Image.Description = "{{ image.distribution|capfirst }} {{ image.release }} {{ image.architecture_mapped }}{% if image.variant != \"default\" %} ({{ image.variant }}){% endif %} ({{ image.serial }})"
 	}
 
-	// Set default target type. This will only be overriden if building VMs for LXD.
+	// Set default target type. This will only be overridden if building VMs for Incus.
 	d.Targets.Type = DefinitionFilterTypeContainer
 }
 
@@ -358,12 +366,14 @@ func (d *Definition) Validate() error {
 		"centos-http",
 		"springdalelinux-http",
 		"debootstrap",
+		"rpmbootstrap",
 		"fedora-http",
 		"gentoo-http",
 		"ubuntu-http",
 		"sabayon-http",
 		"docker-http",
 		"oraclelinux-http",
+		"openeuler-http",
 		"opensuse-http",
 		"openwrt-http",
 		"plamolinux-http",
@@ -371,8 +381,12 @@ func (d *Definition) Validate() error {
 		"funtoo-http",
 		"rootfs-http",
 		"rockylinux-http",
+		"vyos-http",
+		"slackware-http",
+		"nixos-http",
 	}
-	if !shared.StringInSlice(strings.TrimSpace(d.Source.Downloader), validDownloaders) {
+
+	if !util.ValueInSlice(strings.TrimSpace(d.Source.Downloader), validDownloaders) {
 		return fmt.Errorf("source.downloader must be one of %v", validDownloaders)
 	}
 
@@ -390,8 +404,10 @@ func (d *Definition) Validate() error {
 			"xbps",
 			"zypper",
 			"luet",
+			"slackpkg",
 		}
-		if !shared.StringInSlice(strings.TrimSpace(d.Packages.Manager), validManagers) {
+
+		if !util.ValueInSlice(strings.TrimSpace(d.Packages.Manager), validManagers) {
 			return fmt.Errorf("packages.manager must be one of %v", validManagers)
 		}
 
@@ -431,14 +447,13 @@ func (d *Definition) Validate() error {
 		"hostname",
 		"hosts",
 		"remove",
-		"upstart-tty",
 		"cloud-init",
-		"lxd-agent",
+		"incus-agent",
 		"fstab",
 	}
 
 	for _, file := range d.Files {
-		if !shared.StringInSlice(strings.TrimSpace(file.Generator), validGenerators) {
+		if !util.ValueInSlice(strings.TrimSpace(file.Generator), validGenerators) {
 			return fmt.Errorf("files.*.generator must be one of %v", validGenerators)
 		}
 	}
@@ -454,11 +469,12 @@ func (d *Definition) Validate() error {
 		"plamolinux",
 		"voidlinux",
 		"funtoo",
+		"slackware",
 	}
 
 	architectureMap := strings.TrimSpace(d.Mappings.ArchitectureMap)
 	if architectureMap != "" {
-		if !shared.StringInSlice(architectureMap, validMappings) {
+		if !util.ValueInSlice(architectureMap, validMappings) {
 			return fmt.Errorf("mappings.architecture_map must be one of %v", validMappings)
 		}
 	}
@@ -471,7 +487,7 @@ func (d *Definition) Validate() error {
 	}
 
 	for _, action := range d.Actions {
-		if !shared.StringInSlice(action.Trigger, validTriggers) {
+		if !util.ValueInSlice(action.Trigger, validTriggers) {
 			return fmt.Errorf("actions.*.trigger must be one of %v", validTriggers)
 		}
 	}
@@ -482,7 +498,7 @@ func (d *Definition) Validate() error {
 	}
 
 	for _, set := range d.Packages.Sets {
-		if !shared.StringInSlice(set.Action, validPackageActions) {
+		if !util.ValueInSlice(set.Action, validPackageActions) {
 			return fmt.Errorf("packages.*.set.*.action must be one of %v", validPackageActions)
 		}
 	}
@@ -496,19 +512,19 @@ func (d *Definition) Validate() error {
 	d.Image.ArchitectureMapped = archMapped
 
 	// Kernel architecture and personality
-	archID, err := lxdarch.ArchitectureId(d.Image.Architecture)
+	archID, err := incusArch.ArchitectureId(d.Image.Architecture)
 	if err != nil {
 		return fmt.Errorf("Failed to get architecture ID: %w", err)
 	}
 
-	archName, err := lxdarch.ArchitectureName(archID)
+	archName, err := incusArch.ArchitectureName(archID)
 	if err != nil {
 		return fmt.Errorf("Failed to get architecture name: %w", err)
 	}
 
 	d.Image.ArchitectureKernel = archName
 
-	archPersonality, err := lxdarch.ArchitecturePersonality(archID)
+	archPersonality, err := incusArch.ArchitecturePersonality(archID)
 	if err != nil {
 		return fmt.Errorf("Failed to get architecture personality: %w", err)
 	}
@@ -539,7 +555,7 @@ func (d *Definition) GetRunnableActions(trigger string, imageTarget ImageTarget)
 }
 
 // GetEarlyPackages returns a list of packages which are to be installed or removed earlier than the actual package handling
-// Also removes them from the package set so they aren't attempted to be re-installed again as normal packages
+// Also removes them from the package set so they aren't attempted to be re-installed again as normal packages.
 func (d *Definition) GetEarlyPackages(action string) []string {
 	var early []string
 
@@ -599,6 +615,7 @@ func getFieldByTag(v reflect.Value, t reflect.Type, tag string) (reflect.Value,
 			if value < 0 || value >= v.Len() {
 				return reflect.Value{}, errors.New("Index out of range")
 			}
+
 			return getFieldByTag(v.Index(value), t.Elem(), parts[1])
 		}
 
@@ -614,6 +631,7 @@ func getFieldByTag(v reflect.Value, t reflect.Type, tag string) (reflect.Value,
 				if len(parts) == 1 {
 					return v.Field(i), nil
 				}
+
 				return getFieldByTag(v.Field(i), t.Field(i).Type, parts[1])
 			}
 		}
@@ -625,15 +643,15 @@ func getFieldByTag(v reflect.Value, t reflect.Type, tag string) (reflect.Value,
 
 // ApplyFilter returns true if the filter matches.
 func ApplyFilter(filter Filter, release string, architecture string, variant string, targetType DefinitionFilterType, acceptedImageTargets ImageTarget) bool {
-	if len(filter.GetReleases()) > 0 && !shared.StringInSlice(release, filter.GetReleases()) {
+	if len(filter.GetReleases()) > 0 && !util.ValueInSlice(release, filter.GetReleases()) {
 		return false
 	}
 
-	if len(filter.GetArchitectures()) > 0 && !shared.StringInSlice(architecture, filter.GetArchitectures()) {
+	if len(filter.GetArchitectures()) > 0 && !util.ValueInSlice(architecture, filter.GetArchitectures()) {
 		return false
 	}
 
-	if len(filter.GetVariants()) > 0 && !shared.StringInSlice(variant, filter.GetVariants()) {
+	if len(filter.GetVariants()) > 0 && !util.ValueInSlice(variant, filter.GetVariants()) {
 		return false
 	}
 
